/*
 * FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 * Copyright (C) 2005-2014, Anthony Minessale II <anthm@freeswitch.org>
 *
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 *
 * The Initial Developer of the Original Code is
 * Piotr Gregor <piotr@freeswitch.com> <piotr@dataandsignal.com>
 *
 * Portions created by the Initial Developer are Copyright (C)
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Piotr Gregor <piotr@freeswitch.org> <piotr@signalwire.com>
 *
 * mod_red.c -- The redundancy encoder/decoder (RFC 2198, RFC 5109)
 *
 */

#include <switch.h>
#include <switch_red.h>

SWITCH_DECLARE(switch_rtp_hdr_t*) switch_red_gen_red_hdr(switch_rtp_packet_t *rtp, uint8_t red_pt, void *inline_where)
{
	switch_rtp_hdr_t *r = inline_where;

	if (!rtp || !r) {
		return NULL; /* Graceful crash. */
	}

	memcpy(r, &rtp->header, sizeof(switch_rtp_hdr_t));
	/*r->m = 0;*/								/* RFC 5109: 10.3
											   In the payload format for redundant coding as specified by RFC 2198,
											   the marker bit is lost as soon as the primary coding is carried in the RED packets.
											   So the marker bit cannot be recovered regardless of whether or not the FEC is used.

											  RFC 2198: 4  Limitations
											  The RTP marker bit is not preserved for redundant data blocks. Hence if the primary
											  (containing this marker) is lost, the marker is lost. 
											  
											  But Chrome expects to see the marker, so leave the marker set in the header if it was set. */
	r->pt = red_pt;

	return r;
}

/* Note: supports single Redundant Encoding Block.
 * TODO: add support for multiple blocks of redundant encoding. */ 
SWITCH_DECLARE(uint16_t) switch_red_data_len(uint16_t rtp_len, uint16_t redundancy_len)
{
	uint16_t len = 0;
	len += 1 + rtp_len;																					/* Primary Encoding Block Header is 1 octet long. */
	
	if (redundancy_len > 0) {
		len += redundancy_len + sizeof(ulp_red_encoding_block_hdr_t);									/* Redundant Encoding Block Header is complete 4 octets long header. */
	}

	return len;
}

SWITCH_DECLARE(void*) switch_red_gen_red_data(switch_rtp_t *rtp_session, void *rtp, uint16_t rtp_len, uint8_t primary_pt, void *redundancy, uint16_t redundancy_len, uint8_t redundancy_pt, uint16_t red_ts_off_diff, uint16_t *total_data_len, void *inline_where)
{
	switch_core_session_t *session = switch_rtp_get_core_session(rtp_session);
	uint8_t *rd = inline_where;
	uint16_t len = switch_red_data_len(rtp_len, redundancy_len);
	ulp_red_encoding_block_hdr_t *en = NULL;
	
	*total_data_len = 0;
	
	if (!rtp) {
		return NULL; /* Graceful crash. */
	}
	
	if (rd == NULL) {
		rd = switch_core_session_alloc(session, len);
		inline_where = rd;
	}

	en = (ulp_red_encoding_block_hdr_t *) rd;
	
	if (redundancy_len != 0) {
		en->f = 1;                                                              /* This is Redundant Encoding Block Header. */
		en->pt = redundancy_pt;                                                 /* Redundancy's payload type, for ULPFEC this is ULP_PT_ULPFEC. */
		
		ulp_red_encoding_block_set_ts_off_len(en->ts_off_len, red_ts_off_diff, redundancy_len);     /* Produce TS and LEN fields in Redundant Encoding Block Header. This represents diff offset from primary media packet. */
		rd = (uint8_t *) (en + 1);
		memcpy(rd, redundancy, redundancy_len);
		rd += redundancy_len;
	}
	
	*rd = 0x7F & primary_pt;                                                    /* This is Primary Encoding Block Header. It contains only F bit and PT. F is 0 to signal primary encoding, PT is payload type of the RTP media packet. TS and LEN can be deducted from total packet length and RED RTP header. */
	
	rd++;

	memcpy(rd, rtp, rtp_len);
	*total_data_len = len;
	
	return inline_where;
}

SWITCH_DECLARE(switch_rtp_packet_t*) switch_red_gen_pkt(switch_rtp_t* rtp_session, void *rtp, uint16_t rtp_len, uint8_t rtp_pt, uint8_t clean_hdr, void *redundancy, uint16_t redundancy_len, uint8_t redundancy_pt, uint16_t red_ts_off_diff, uint16_t *total_len, uint8_t red_pt, void *inline_where)
{
	switch_rtp_packet_t *red = inline_where;
	switch_core_session_t *session = switch_rtp_get_core_session(rtp_session);


	*total_len = 0;

	if (!rtp || (!red && !session)) {
		return NULL;																/* Graceful crash */
	}

	if (!clean_hdr) {																/* RED RTP header */
		if (!switch_red_gen_red_hdr(rtp, red_pt, red)) {
			return NULL;
		}
		/* Optional Redundant Encoding Block (encoding hdr + redunancy) + Primary Encoding Block (encoding header + RTP media packet). */
		rtp = (char *) rtp + sizeof(switch_rtp_hdr_t);
	} else {
		if (red) {
			memset(red, 0, sizeof(switch_rtp_hdr_t));
		}
	}

	if (!switch_red_gen_red_data(rtp_session, rtp, rtp_len, rtp_pt, redundancy, redundancy_len, redundancy_pt, red_ts_off_diff, total_len, (char *) red + sizeof(switch_rtp_hdr_t))) {
		return NULL;																/* Graceful crash */
	}

	*total_len += sizeof(switch_rtp_hdr_t);

	return red;
}

SWITCH_DECLARE(switch_rtp_packet_t*) switch_red_gen_pkt_clean_header(switch_rtp_t *rtp_session, void *rtp, uint16_t rtp_len, uint8_t rtp_pt, void *redundancy, uint16_t redundancy_len, uint8_t redundancy_pt, uint16_t red_ts_off_diff, uint16_t *total_len, uint8_t red_pt, void *inline_where)
{
	return switch_red_gen_pkt(rtp_session, rtp, rtp_len, rtp_pt, 1, redundancy, redundancy_len, redundancy_pt, red_ts_off_diff, total_len, red_pt, inline_where);
}

SWITCH_DECLARE(switch_rtp_packet_t*) switch_red_gen_pkt_with_header(switch_rtp_t *rtp_session, void *rtp, uint16_t rtp_len, uint8_t rtp_pt, void *redundancy, uint16_t redundancy_len, uint8_t redundancy_pt, uint16_t red_ts_off_diff, uint16_t *total_len, uint8_t red_pt, void *inline_where)
{
	return switch_red_gen_pkt(rtp_session, rtp, rtp_len, rtp_pt, 0, redundancy, redundancy_len, redundancy_pt, red_ts_off_diff, total_len, red_pt, inline_where);
}

SWITCH_DECLARE(switch_rtp_packet_t*) switch_red_depacketize(switch_core_session_t* session, ulp_policy_t *ulp_policy, switch_rtp_hdr_t *rtp, uint16_t *rtp_len, void **redundancy, uint16_t *redundancy_len, uint16_t *red_ts_off_diff, void* inline_where, uint8_t ulpfec_pt)
{
	uint8_t *rd = NULL;
	ulp_red_encoding_block_hdr_t *en = NULL;
	uint16_t red_fec_ts_diff_offset = 0;
	uint16_t red_fec_len = 0;

	char *prim_encoding = NULL;
	uint8_t prim_encoding_pt = 0;
	uint16_t prim_encoding_len = 0;
	switch_rtp_packet_t *rtp_out = NULL;

	*red_ts_off_diff = 0;
	*redundancy_len = 0;


	if (!rtp || (*rtp_len < 1)) {												/* There must be at least Primary Encoding Block Header in the RED payload. */
		return NULL;
	}

	rd = (uint8_t *) (rtp + 1);


	/* Packet validation. */

	if (*rd & 0x80) {															/* Check if this is Redundant Encoding header. */
		en = (void *) rd;														/* Indicate Redundancy presence in the packet. */

		red_fec_ts_diff_offset = ulp_red_encoding_block_get_ts(en->ts_off_len);
		red_fec_len = ulp_red_encoding_block_get_len(en->ts_off_len);
#if DEBUG_RED
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "RED: found Redundant Encoding of [%u] bytes\n", red_fec_len);
#endif 
		rd = (uint8_t *) en + sizeof(ulp_red_encoding_block_hdr_t) + red_fec_len;	/* Move on to the Primary Encoding Block Header. */
	}

	if (*rd & 0x80) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_WARNING, "RED: multiple redundant encodings unsupported.\n");
		return NULL;															/* We do not support multiple redundant encodings. TODO: add support for multiple encodings. */
	}

	prim_encoding_pt = *rd & 0x7f;

	if (prim_encoding_pt == ulpfec_pt) {
		
		/* Chromium! */
		if (ulp_policy) {
			ulp_set_flag(ulp_policy, ULP_FL_REDUNDANT_AS_PRIMARY);
		}

#if DEBUG_RED
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "RED: Redundant sent as Primary Encoding (probably Chromium! ...)\n");
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "RED: len=%u\n", *rtp_len);
#endif
	}
	
	prim_encoding = (char *) rd + 1;											/* +1 as good books say... */

	if ((en && sizeof(ulp_red_encoding_block_hdr_t) + red_fec_len >= *rtp_len) || 
			(red_fec_len > SWITCH_RTP_MAX_BUF_LEN - sizeof(ulp_rtp_hdr_t) - 1 /*RED header size included*/)) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_WARNING, "RED: invalid length for Redundant Encoding.\n");
		return NULL;
	}

	prim_encoding_len = *rtp_len - ((en != NULL ? sizeof(ulp_red_encoding_block_hdr_t) + red_fec_len : 0) + 1);

#if DEBUG_RED
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "RED: len=%u plen=%u\n", *rtp_len, prim_encoding_len);
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "RED: Primary Encoding header %02x and then ... %02x %02x %02x\n", *rd, *(rd + 1), *(rd + 2), *(rd + 3));
#endif

	/* 
	 * RFC 5109 10.3
	 *
	 * The conversion from RED packets to virtual RTP packets is simply done by
	 * (1) removing any RED block headers and redundant coding data,
	 * and (2) replacing the PT in the RTP header with the PT of the primary coding.
	 */

	if (en) {																	/* Check if this is Redundant Encoding header. */

		if ((red_fec_len > 0) && (! *redundancy)) {
			 switch_malloc(*redundancy, red_fec_len);
		}
	}

	if (en) {																	/* Check if this is Redundant Encoding header. */

		if (red_fec_len > 0) {
			memcpy(*redundancy, (void *) (en + 1), red_fec_len);
		}

		*red_ts_off_diff = red_fec_ts_diff_offset;
		*redundancy_len = red_fec_len;
	}

#if DEBUG_RED
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "RED: len=%u plen=%u\n", *rtp_len, prim_encoding_len);
#endif

	memcpy(inline_where, rtp, sizeof(ulp_rtp_hdr_t));
	memcpy((char *) inline_where + sizeof(ulp_rtp_hdr_t), prim_encoding, prim_encoding_len);
	rtp_out = inline_where;
	rtp_out->header.pt = prim_encoding_pt;
	*rtp_len = prim_encoding_len;

#if DEBUG_RED
	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "RED: len=%u plen=%u\n", *rtp_len, prim_encoding_len);
#endif

	return rtp_out;
}

/**
 * Return 1 - valid, 0 - invalid.
 */
uint8_t ulp_cb_validate_recovery(ulp_rtp_pkt_t *new_rtp, ulp_rtp_msg_t *rtp, uint32_t rtp_n)
{
	// todo, always valid for now 
	return 1;
}

SWITCH_DECLARE(uint8_t) switch_rtp_recover(ulp_policy_t *ulp_policy, uint16_t seq, ulp_status_t *ulp_err, switch_rtp_packet_t *recovered_pkt, uint32_t ssrc)
{
	ulp_rtp_pkt_t *rtp = NULL;
	uint8_t recovered_packet_valid = 0;

	if (!ulp_policy) {
		return 0;
	}

	/* @recovered_packet_valid is set to 1 if recovery succeeded for the final packet returned from @ulp_recover.
	 * @ulp_recover may call @ulp_cb_validate_recovery many times - once for each recovery-candidate packet. */

	rtp = ulp_recover(ulp_policy, seq, ulp_err, &recovered_packet_valid, ulp_cb_validate_recovery);
	if (!recovered_packet_valid) {
		if (rtp) {
			free(rtp);
		}
		return 0;
	}

	if (rtp) {
		memset(recovered_pkt, 0, sizeof(switch_rtp_packet_t));
		memcpy(recovered_pkt, rtp, ulp_rtp_last_recover_len(ulp_policy));
		recovered_pkt->header.ssrc = htonl(ssrc);
	}

#if 0
#ifdef TEST_ULPFEC
	if (*ulp_err == ulp_status_ok) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "RTP FEC RED: TEST: Recovery for seq=%u OK\n", seq);
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "RTP FEC RED: TEST: Recovery validation=%u\n", recovered_packet_valid);
	}
#endif // TEST_ULPFEC
#endif 

	if (rtp) {
		free(rtp);
	}

	return rtp ? 1 : 0;
}

SWITCH_DECLARE(switch_status_t) switch_red_ulpfec_parse(ulp_policy_t *ulp_policy, void *redundancy, uint16_t redundancy_len)
{
	ulp_fec_lvl_64_info_t lvl[SWITCH_ULP_MAX_LVL_N];
	uint8_t lvl_n = SWITCH_ULP_MAX_LVL_N;
	uint8_t more = 0;
	uint8_t l = 0;


	memset(lvl, 0, sizeof(lvl));

	if (!redundancy || (redundancy_len < sizeof(ulp_fec_hdr_t))) {
		return SWITCH_STATUS_FALSE;
	}

	if (ulp_fec_parse(ulp_policy, redundancy, redundancy_len, lvl, &lvl_n,  &more, &l) != ulp_status_ok) {
#if DEBUG_RED
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "ULPFEC policy parsing failed: ulp_fec_parse\n");
#endif
		return SWITCH_STATUS_FALSE;
	}
	
	if (lvl_n == 0) {
#if DEBUG_RED
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "ULPFEC policy: no protection level headers, lvl_n=0\n");
#endif
		return SWITCH_STATUS_NOTFOUND;
	}

	if (ulp_policy) {
		
		memset(ulp_policy->lvl, 0, sizeof(ulp_policy->lvl));

		ulp_policy->lvl_n = lvl_n;

		while (lvl_n > 0) {
			--lvl_n;
			ulp_policy->lvl[lvl_n].lk = lvl[lvl_n].lk;
			ulp_policy->lvl[lvl_n].gk = lvl[lvl_n].gk;

			ulp_policy->lvl[lvl_n].mask16 = lvl[lvl_n].mask16;
			ulp_policy->lvl[lvl_n].mask_cont = lvl[lvl_n].mask_cont;

#if DEBUG_RED
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_INFO, "ULPFEC policy: k=%u lk=%u gk=%u mask16=%u[%.4x] mask_cont=%u[%.8x]\n", lvl_n, lvl[lvl_n].lk, lvl[lvl_n].gk, lvl[lvl_n].mask16, lvl[lvl_n].mask16, lvl[lvl_n].mask_cont, lvl[lvl_n].mask_cont);
#endif
		}
	}

	if (more) {
#if DEBUG_RED
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "ULPFEC policy parsing failed: level lengths do not match the size of the FEC\n");
#endif
		return SWITCH_STATUS_MORE_DATA;
	}


	return SWITCH_STATUS_SUCCESS;
}
